from . import JobDAO, Connection, JobStatus, JobAction, JobCreator as JC
import threading

__author__ = 'rodak'

'''
How to use a connection to database:

try:
   db = database.getDB()
   dao = database.dao(db)
   ...
finally:
   database.releaseDB(db)

Always release db if you don't need it anymore.

'''

dbData = threading.local()

def getDB():
    if "nr" not in dbData.__dict__:
        dbData.nr = 0
    if dbData.nr == 0:
        dbData.db = Connection.createDB()
    dbData.nr += 1
    return dbData.db

def releaseDB(db):
    dbData.nr -= 1
    if dbData.nr == 0:
        dbData.db.connection.close()
    if dbData.nr < 0:
        raise Exception("Inconsistent use of getDB and releaseDB.")

def dao(db):
    return JobDAO.JobDAO(db)

def context(handler):
    db = getDB()
    try:
        handler(db)
    finally:
        releaseDB(db)

JobDependency = JobDAO.Dependency
JobOption = JobDAO.Option
JobCreator = JC.JobCreator


jobStatus = JobStatus
jobAction = JobAction


'''
Created on Jun 5, 2014

@author: rodak
'''
from . import Native

from . import Executer
from . import Logging
from Jobber import Settings

logger = Logging.logger("LocalJobExecuter", Settings.logDir+"/LocalJobExecuter.log")

'''
An executer that will run jobs locally
'''

class JobExecuter(Executer.Executer):

    def __init__(self, response):
        Executer.Executer.__init__(self, response)
        self.pooledExecuter = Native.SystemCommandPooledExecuter()

    def start(self):
        Executer.Executer.start(self)
        self.pooledExecuter.start()
        logger.info("LocalJobExecuter Started")

    def onStop(self):
        self.pooledExecuter.stop()
        logger.info("LocalJobExecuter Stopped")

    def onExecute(self, job, script, logDir):
        logger.info("Execute job "+str(job.jobId))
        runJobMess = Native.RunJobMessage(script, logDir+"/out", logDir+"/err", lambda res: self.processCommandResult(res))
        runJobMess.maxRuntime = 60*60*24
        lastId = self.pooledExecuter.execute(runJobMess)
        self.response.executionId(job.jobId, lastId)

    def processCommandResult(self, result):
        error = None
        if result.status == Native.TIMEOUT:
            error = "Job timed out"
        elif result.status == Native.KILLED:
            error = "Job was killed"
        else:
            if result.returnCode != 0:
                error = "Job exited with exit status "+str(result.returnCode)
        self.response.finish(result.jobId, error)

    def onTerminate(self, executionId):
        logger.info("Kill execution "+executionId)
        self.pooledExecuter.kill(executionId)

    def onStatus(self, executionId):
        self.response.status(executionId, "unknown")


import uuid

from Jobber import Executer


__author__ = 'rodak'

'''
A simple executer that just reports a job as successfully executed without executing it. For testing purposes.
'''

class JobExecuter(Executer.Executer):

    def __init__(self, response):
        Executer.Executer.__init__(response)

    def onExecute(self, job, script, logDir):
        exId = str(uuid.uuid4())
        self.response.executionId(job.jobId, exId)
        self.response.finish(exId, None)

    def onStatus(self, executionId):
        self.response.status(executionId, "unknown")

    def onTerminate(self, executionId):
        pass


__author__ = 'rodak'

import logging

def logger(name, outputPath):
    logger = logging.getLogger(name)
    handler = logging.FileHandler(outputPath)
    handler.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
    logger.addHandler(handler)
    return logger


import os

__author__ = 'rodak'

import sqlite3

from Jobber import Settings

def connect():
    return sqlite3.connect(Settings.sqlite_path, isolation_level=None, check_same_thread=False)

def init(db):
    f = open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "JobDatabaseSQLite.sql"))
    str = f.read()
    db.connection.executescript(str)
    f.close()

def dict_factory(cursor, row):
    d = {}
    for idx,col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d

class DB():

    def __init__(self):
        self.connection = connect()
        self.connection.row_factory = dict_factory
        self.cursor = self.connection.cursor()

    def execute(self, query, params=[]):
        query = query.replace("%s", "?")
        try:
            self.cursor.execute(query, params)
        except (AttributeError, sqlite3.OperationalError):
            self.connection = connect()
            self.connection.row_factory = dict_factory
            self.cursor = self.connection.cursor()
            self.cursor.execute(query, params)
        return self.cursor

    def insert_id(self):
        return self.cursor.lastrowid

    def fetchall(self):
        return self.cursor.fetchall()

    def fetchone(self):
        return self.cursor.fetchone()

    def close(self):
        self.connection.close()

    def init(self):
        init(self)


import sys
sys.path.append("..")
from Jobber import database
import graphviz as gv
from flask import Flask, render_template, url_for, redirect

g2 = gv.Digraph(format='svg')
db = database.getDB()
for d in db.getMembers(int(sys.argv[1])):
    print d['status'], d['name'], d['is_group_job']
    g2.node(d['name'])

for d in db.getMembers(int(sys.argv[1])):
    for dep in db.getDependentJobs(d['id']):
        g2.edge(d['name'], db.getJobInfo(dep)['name'])

g2.render('img/g2')



'''
Created on Jun 5, 2014

@author: rodak
'''
import threading
from Queue import Queue, Empty
from time import sleep

import drmaa
from drmaa.errors import ExitTimeoutException, InvalidJobException
import math
from . import Settings

from Jobber import Executer, Logging

logger = Logging.logger("DRMAAJobExecuter", Settings.logDir+"/DRMAAJobExecuter.log")

maxNrOfJobs = Settings.drmaaMaxNrOfJobs

'''
An executer that will run jobs on a drmaa system
'''

class JobExecuter(Executer.Executer):

    def __init__(self, response):
        self.session = None
        self.nrOfRunning = 0
        Executer.Executer.__init__(self, response)
        self.jobQueue = []

    def onTerminate(self, executionId):
        self.session.control(str(executionId), drmaa.JobControlAction.TERMINATE)

    def onStatus(self, executionId):
        try:
            jobStatus = self.session.jobStatus(executionId)
        except InvalidJobException:
            jobStatus = "unknown"
        self.response.status(executionId, jobStatus)

    def onExecute(self, job, script, logDir):
        logger.info("Received message execute job "+str(job.jobId))
        if self.nrOfRunning < maxNrOfJobs:
            self._execJob(job, script, logDir)
        else:
            self.jobQueue.append((job, script, logDir))

    def _execJob(self, job, script, logDir):
        jobTemplate = self.transformJobToDRMAAJobTemplate(job, script, logDir)
        djobId = self.session.runJob(jobTemplate)
        logger.info("Executed job "+str(job.jobId)+" with execution id "+str(djobId))
        self.response.executionId(job.jobId, str(djobId))
        self.nrOfRunning += 1

    def _getJobError(self, val):
        if val.hasExited and val.exitStatus == 0:
            return None
        else:
            if val.hasExited:
                return "Failed because of exit value "+str(val.exitStatus)
            elif val.hasSignal:
                return "Failed because received signal "+str(val.terminatedSignal)
            elif val.wasAborted:
                return "Failed because was aborted"
            elif val.hasCoreDump:
                return "Failed because of core dump"
            else:
                return "Failed because of unknown error"

    def onRun(self):
        try:
            nr = 0
            while nr < 100:
                nr += 1
                retval = self.session.wait(drmaa.Session.JOB_IDS_SESSION_ANY, drmaa.Session.TIMEOUT_NO_WAIT)
                self.nrOfRunning -= 1
                while self.nrOfRunning < maxNrOfJobs and len(self.jobQueue) > 0:
                    mess = self.jobQueue.pop(0)
                    self._execJob(mess[0], mess[1], mess[2])
                logger.info("Job finished "+str(retval.jobId))
                self.response.finish(str(retval.jobId), self._getJobError(retval))
        except ExitTimeoutException:
            pass
        except InvalidJobException:
            pass
        except Exception, e:
            logger.warn(str(e))
            if e.message.startswith("code 24: no usage information was returned for the completed job"):
                pass
            else:
                logger.exception(str(e))

    def run(self):
        try:
            logger.info("started drmaa job executer")
            self.session = drmaa.Session()
            self.session.initialize()
            Executer.Executer.run(self)
        finally:
            self.session.exit()
            logger.info("drmaa job executer stopped")

    def transformJobToDRMAAJobTemplate(self, job, script, logDir):
        jt = self.session.createJobTemplate()
        nativeSpec = "-w n"
        nrOfCores = 1
        for opt in job.options:
            if opt.name == "cores":
                nrOfCores = opt.value
        for opt in job.options:
            if opt.name == "memory":
                val = opt.value
                memByCore = float(val[:-1]) / float(nrOfCores)
                memByCore = int(math.ceil(memByCore))
                nativeSpec += " -l membycore="+str(memByCore)+val[len(val) - 1]
            elif opt.name == "runtime":
                nativeSpec += " -l runtime="+opt.value
            elif opt.name == "cores":
                nativeSpec += " -pe smp "+opt.value
            elif not opt.name == 'module':
                nativeSpec += " -"+opt.name+" "+opt.value
        jt.nativeSpecification = nativeSpec
        jt.remoteCommand = script
        jt.outputPath = ":"+logDir+"/out"
        jt.errorPath = ":"+logDir+"/err"
        jt.jobName = job.name
        return jt


import Queue
from Queue import Empty
import os
from Jobber import Logging, Settings, JobExecuter

'''
Usage:

executer = Executers.create(messageHandler)
executer.run(job)

where:
    messageHandler an object of type MessageHandler
    job is a database job. see class "JobCreator" in module "database".

'''


__author__ = 'rodak'
executers = {}

if "JOBBER_EXECUTERS" in os.environ:
    execfile(os.environ["JOBBER_EXECUTERS"])
else:
    execfile(os.path.join(os.path.expanduser("~"), ".jobber")+"/executers.py")

logger = Logging.logger("Executers", Settings.logDir+"/Executers.log")

def _createExecuters(listener):
    idToExecuter = {};
    for name in executers.keys():
        executer = executers[name]
        idToExecuter[name] = executer(listener)
        idToExecuter[name].start()
        logger.info("Loaded executer "+name)
    return idToExecuter


class ResponseHandler():

    def __init__(self):
        self.queue = Queue.Queue()
        self.batchNr = 100

    def _addMessage(self, message):
        self.queue.put(message)

    def executionId(self, jobId, executerId):
        self.queue.put(("execution", jobId, executerId))

    def finish(self, executerId, error = None):
        self.queue.put(("finish", executerId, error))

    def status(self, executionId, jobStatus):
        self.queue.put(("status", executionId, jobStatus))

    def onExecutionId(self, jobId, executerId):
        raise Exception("onUpdateId must be implemented by subclasses")

    def onFinish(self, executerId, error):
        raise Exception("onJobFinished must be implemented by subclasses")

    def onStatus(self, jobId, status):
        pass

    def handleMessages(self):
        try:
            nr = 0
            while nr < self.batchNr:
                nr += 1
                mess = self.queue.get(False)
                if mess[0] == "execution":
                    self.onExecutionId(mess[1], mess[2])
                elif mess[0] == "finish":
                    self.onJobFinished(mess[1], mess[2])
                elif mess[0] == "status":
                    self.onStatus(mess[1], mess[2])
        except Empty:
            pass

class MultiExecuter():

    def __init__(self, messageHandler):
        self.idToExecuter = _createExecuters(messageHandler)
        self.messageHandler = messageHandler

    def run(self, job):
        logger.info("executing job "+str(job.jobId))
        logDir = JobExecuter.createLogDir(job)
        script = JobExecuter.createRunScript(job)
        ex = job.executer
        if job.executer not in self.idToExecuter:
            logger.warn("Executer with id "+ex+" does not exist for job with id "+str(job.jobId)+", falling back to executer "+Settings.defaultExecuter)
            ex = Settings.defaultExecuter
        self.idToExecuter[ex].execute(job, script, logDir)

    def handleMessages(self):
        self.messageHandler.handleMessages()

    def terminate(self, executer, executionId):
        if executer not in self.idToExecuter:
            logger.warn("Executer with id "+executer+" does not exist for job with id "+str(job.jobId)+", falling back to executer "+Settings.defaultExecuter)
            executer = Settings.defaultExecuter
        self.idToExecuter[executer].terminate(executionId)



    def stop(self):
        for ex in self.idToExecuter:
            self.idToExecuter[ex].stop()

def create(handler):
    return MultiExecuter(handler)



import Queue
from Queue import Empty
import threading
from time import sleep
from Jobber import Settings, Logging

__author__ = 'rodak'


logger = Logging.logger("Executer", Settings.logDir+"/Executer.log")

class Executer(threading.Thread):

    def __init__(self, response):
        threading.Thread.__init__(self)
        self.response = response
        self.queue = Queue.Queue()
        self.batchNr = 100

    def terminate(self, executerId):
        self.queue.put(("terminate", executerId))

    def execute(self, job, script, logDir):
        self.queue.put(("execute", job, script, logDir))

    def status(self, executerId):
        self.queue.put("status", executerId)

    def stop(self):
        self.queue.put(["stop"])

    def onTerminate(self, jobId):
        raise Exception("onTerminate must be implemented by subclasses")

    def onExecute(self, job, script, logDir):
        raise Exception("onExecute must be implemented by subclasses")

    def onStatus(self):
        raise Exception("onStatus must be implemented by subclasses")

    def onStop(self):
        pass

    def onRun(self):
        pass

    def run(self):
        running = True
        while running:
            try:
                try:
                    nr = 0
                    while nr < self.batchNr:
                        nr += 1
                        mess = self.queue.get(False)
                        if mess[0] == "terminate":
                            self.onTerminate(mess[1])
                        elif mess[0] == "execute":
                            self.onExecute(mess[1], mess[2], mess[3])
                        elif mess[0] == "stop":
                            running = False
                            self.onStop()
                            return
                except Empty:
                    pass
                self.onRun()
            except Exception, e:
                logger.exception(str(e))
            sleep(0.2)



from Jobber.database import JobStatus


class Job():
    def __init__(self):
        pass

class JobCreator():

    def __init__(self):
        self.outputPath = None
        self.reset()


    def reset(self):
        self.maxNrOfParallelJobs = 0
        self.maxNrOfRestarts = 3
        self.name = "anonymous"
        self.description = "none"
        self.uniqueJob = False
        self.uniqueKey = None
        self.groupJob = False
        self.dependencies = []
        self.options = []
        self.members = []
        self.deleteTime = 0
        self.executer = None

    def createJob(self, command):
        job = Job()
        job.name = self.name
        job.description = self.description
        job.status = JobStatus.DEACTIVATED
        job.command = command
        job.unique = self.uniqueJob
        job.uniqueKey = self.uniqueKey
        job.groupJob = self.groupJob
        job.startDate = None
        job.endDate = None
        job.maxParallelNr = self.maxNrOfParallelJobs
        job.maxNrOfRestarts = self.maxNrOfRestarts
        job.currentRun = 0
        job.deleteTime = self.deleteTime
        job.whoCreate = None
        job.whoUpdate = None
        job.options = self.options
        job.dependencies = self.dependencies
        job.members = self.members
        job.parentId = None
        job.executer = self.executer
        self.reset()
        return job


import MySQLdb
import os

from Jobber import Settings

__author__ = 'rodak'

def run_sql_file(filename, connection):
    file = open(filename, 'r')
    sql = " ".join(file.readlines())
    lines = sql.split(";")
    for line in lines:
        line = line.strip()
        if line != "":
            cursor = connection.cursor()
            cursor.execute(line)
            connection.commit()
            cursor.close()

def connect():
    return MySQLdb.connect(
        host = Settings.db_host,
        user = Settings.db_user,
        passwd = Settings.db_password,
        db = Settings.db_name)

def init(db):
    run_sql_file(os.path.join(os.path.dirname(os.path.realpath(__file__)), "JobDatabase.sql"), db.connection)


class DB():

    def __init__(self):
        self.connection = connect()
        self.cursor = self.connection.cursor(MySQLdb.cursors.DictCursor)

    def execute(self, query, params=[]):
        try:
            self.cursor.execute(query, params)
        except (AttributeError, MySQLdb.OperationalError):
            self.connection = connect()
            self.cursor = self.connection.cursor(MySQLdb.cursors.DictCursor)
            self.cursor.execute(query, params)
        return self.cursor

    def insert_id(self):
        return self.connection.insert_id()

    def fetchall(self):
        return self.cursor.fetchall()

    def fetchone(self):
        return self.cursor.fetchone()

    def close(self):
        self.connection.close()

    def init(self):
        init(self)


import os
from . import Settings

__author__ = 'rodak'


JOB_FINISHED = "Finished"

class JobAndScript():

    def __init__(self, job, script, logDir):
        self.job = job
        self.script = script
        self.logDir = logDir

class DRMAAResponse():

    def __init__(self, messType, message):
        self.type = messType
        self.message = message

class JobIdParts():

    def __init__(self, first, second, third, jobId):
        self.first = first
        self.second = second
        self.third = third
        self.jobId = jobId

    def getPath(self):
        return self.first+"/"+self.second+"/"+self.third

    def __str__(self):
        return self.getPath()

def jobIdToParts(jobId):
    sid = str(jobId)
    if len(sid) > 3:
        if len(sid) > 6:
            third = sid[-3:]
            second = sid[-6: -3]
            first = sid[:len(sid) - 6]
            return JobIdParts(first, second, third, jobId)
        else:
            third = sid[-3:]
            second = sid[:len(sid) - 3]
            return JobIdParts("0", second, third, jobId)
    else:
        return JobIdParts("0", "0", sid, jobId)

def createLogDir(job):
    parts = jobIdToParts(job.jobId)
    logDir = Settings.logDir+"/"+parts.getPath()+"/"+str(job.currentRun)
    if not os.path.exists(logDir):
        os.makedirs(logDir)
    return logDir

def readStartEndRun(job):
    logDir = createLogDir(job)
    if not os.path.exists(logDir+"/start"):
        return None
    if not os.path.exists(logDir+"/end"):
        return None
    with open(logDir+"/start") as f:
        try:
            start = long(f.read())
        except:
            return None
    with open(logDir+"/end") as f:
        try:
            end = long(f.read())
        except:
            return None
    return {
        'start': start,
        'end': end
    }

def createRunScript(job):
    modules = []
    for opt in job.options:
        if opt.name == 'module':
            modules.append(opt.value)
    logDir = createLogDir(job)
    w = open(logDir+"/command.sh", "w")
    w.write("#!/bin/bash\n\n")
    for mod in modules:
        w.write("module load "+mod+"\n")
    w.write("\n")
    w.write("date +%s > "+logDir+"/start\n")
    w.write(job.command+"\n")
    w.write("exitStatus=$?\n")
    w.write("date +%s > "+logDir+"/end\n")
    w.write("exit ${exitStatus}\n")
    w.close()
    os.chmod(logDir+"/command.sh", 0777)
    return logDir+"/command.sh"



'''
Created on May 12, 2014

@author: rodak
'''
import logging

import threading
from Queue import Queue, Empty
import datetime
from time import sleep
import traceback
from uuid import uuid4
import time
import os

from . import database
from . import JobExecuter, Executers, Settings
from . import Logging

loggerJobController = Logging.logger("JobController", Settings.logDir+"/JobController.log")
loggerMessageHandler = Logging.logger("Message handler", Settings.logDir+"/MessageHandler.log")

class ExecutorMessageHandler(Executers.ResponseHandler):

    def __init__(self, controller):
        Executers.ResponseHandler.__init__(self)
        self.controller = controller

    def onExecutionId(self, jobId, executerId):
        loggerJobController.info("Job with id "+str(jobId)+" assigned execution id "+str(executerId))
        database.context(lambda db: database.dao(db).updateDRMAAId(jobId, executerId))


    def onJobFinished(self, executerId, error):
        db = database.getDB()
        dao = database.dao(db)
        try:
            drmaaId = executerId
            job = dao.getJobWithDrmaaId(drmaaId)
            loggerJobController.info("Job with execution id "+str(drmaaId)+" finished")
            if job is not None and job.status != database.jobStatus.CANCELLED:
                jobId = job.jobId
                loggerJobController.info("Job with id "+str(jobId)+" finished")
                status = database.jobStatus.FAILED
                dao.updateEndSubmitDate(jobId, datetime.datetime.now())
                startEnd = JobExecuter.readStartEndRun(job)
                if startEnd is not None:
                    dao.updateStartRunDate(jobId, datetime.datetime.fromtimestamp(startEnd['start']))
                    dao.updateEndRunDate(jobId, datetime.datetime.fromtimestamp(startEnd['end']))
                if error is None:
                    status = database.jobStatus.FINISHED
                    dao.updateError(jobId, "")
                else:
                    dao.updateError(jobId, error)
                if status == database.jobStatus.FAILED:
                    if dao.isRetryJob(jobId):
                        dao.updateJobStatus(jobId, database.jobStatus.IDLE)
                        self.controller.jobs.add(jobId)
                        self.controller.processActiveJobs()
                        return
                dao.updateJobStatus(jobId, status)
                self.controller.addAffectedJobs(jobId)
                self.controller.processActiveJobs()
                if job.parentId is None:
                    activable = dao.getActivableJob(job.name)
                    if activable is not None:
                        self.controller.launchJob(activable)
            else:
                if job is None:
                    loggerJobController.warn("Job with execution id "+str(drmaaId)+" not found")
        finally:
            database.releaseDB(db)


class JobManagerProcess(threading.Thread):

    def registerForUpdates(self, job):
        self.updateableJobs.append(job)

    def __init__(self):
        threading.Thread.__init__(self)
        self.processing = False
        self.daemon = True
        self.queue = Queue()
        self.groups = set()
        self.jobDAO = None
        self.running = True
        self.executer = Executers.create(ExecutorMessageHandler(self))

    def processActiveJobs(self):
        db = database.getDB()
        dao = database.dao(db)
        try:
            if self.processing:
                return
            self.processing = True
            while (len(self.jobs) > 0) or (len(self.groups) > 0):
                active = sorted(self.jobs)
                self.jobs = set([])
                for jobId in active:
                    self.chooseAction(jobId)
                    parent = dao.getParentId(jobId)
                    if parent is not None:
                        self.groups.add(parent)
                    else:
                        name = dao.getJobInfo(jobId)['name']
                        activable = dao.getActivableJob(name)
                        if activable is not None:
                            self.launchJob(activable)
                newGroups = set([])
                for gid in self.groups:
                    action = dao.getGroupJobAction(gid)
                    if action == database.jobAction.CANCEL:
                        dao.updateJobStatus(gid, database.jobStatus.CANCELLED)
                    elif action == database.jobAction.FAIL:
                        dao.updateJobStatus(gid, database.jobStatus.FAILED)
                    elif action == database.jobAction.FINISH_GROUP:
                        dao.updateJobStatus(gid, database.jobStatus.FINISHED)
                        for jid in dao.getAffectedJobs(gid):
                            self.jobs.add(jid)
                    elif action == database.jobAction.EXECUTE:
                        dao.updateJobStatus(gid, database.jobStatus.RUNNING)
                    elif action == database.jobAction.IDLE_GROUP:
                        dao.updateJobStatus(gid, database.jobStatus.IDLE)
                    if action != database.jobAction.NOTHING:
                        parent = dao.getParentId(gid)
                        if parent is not None:
                            newGroups.add(parent)
                        else:
                            name = dao.getJobInfo(gid)['name']
                            activable = dao.getActivableJob(name)
                            if activable is not None:
                                self.launchJob(activable)
                self.groups = newGroups
            self.processing = False
        finally:
            database.releaseDB(db)

    def stop(self):
        self.running = False
        self.queue.put(["stop"])
        self.executer.stop()

    def addAffectedJobs(self, jobId):
        db = database.getDB()
        dao = database.dao(db)
        try:
            for id in dao.getAffectedJobs(jobId):
                self.jobs.add(id)
            parent = dao.getParentId(jobId)
            if parent is not None:
                self.groups.add(parent)
        finally:
            database.releaseDB(db)

    def chooseAction(self, jobId):
        db = database.getDB()
        dao = database.dao(db)
        try:
            action = dao.getJobAction(jobId)
            if action == database.jobAction.EXECUTE:
                job = dao.getJobWithId(jobId)
                dao.updateStartSubmitDate(job.jobId, datetime.datetime.now())
                dao.updateJobStatus(job.jobId, database.jobStatus.RUNNING)
                dao.updateRun(job.jobId)
                job.currentRun += 1
                self.executer.run(job)
        finally:
            database.releaseDB(db)

    def sendMessage(self, mess):
        self.queue.put(mess)

    def updateStatus(self, jobIds, status, recursive, resonseId):
        self.queue.put(("update", jobIds, status, recursive, resonseId))

    def onUpdateStatus(self, jobIds, status, recursive, resonseId):
        db = database.getDB()
        dao = database.dao(db)
        try:
            for jobId in jobIds:
                loggerJobController.info("Received message set job status (job_id, status, recursive): "+str(jobId)+" "+str(status)+" "+str(recursive))
                currentStatus = dao.getJobStatus(jobId)
                groupMembers = True
                if currentStatus != status:
                    parentId = jobId
                    while parentId is not None:
                        affectedJobs = self._updateJobStatus(parentId, status, recursive, groupMembers)
                        for jId in affectedJobs:
                            self.jobs.add(jId)
                        if recursive:
                            parentId = dao.getParentId(parentId)
                        else:
                            parentId = None
                        groupMembers = False
            self.processActiveJobs()
            if resonseId is not None:
                dao.sendResponse(resonseId, True, "ok")
        except Exception, e:
            loggerJobController.exception(str(e))
            if resonseId is not None:
                dao.sendResponse(resonseId, False, e.message)
            raise e
        finally:
            database.releaseDB(db)

    def deleteJob(self, jobId, responseId):
        self.queue.put(("delete", jobId, responseId))

    def onDeleteJob(self, jobId, responseId):
        db = database.getDB()
        dao = database.dao(db)
        try:
            loggerJobController.info("Received message delete job "+str(jobId))
            j = dao.getJobWithId(jobId)
            if j.drmaaId is not None:
                self.executer.terminate(j.executer, j.drmaaId)
                dao.updateDRMAAId(id, None)
            dao.deleteJob(jobId)
            if responseId is not None:
                dao.sendResponse(responseId, True, "ok")
        except Exception, e:
            loggerJobController.exception(str(e))
            if responseId is not None:
                dao.sendResponse(responseId, False, e.message)
            raise e
        finally:
            database.releaseDB(db)

    def launchJob(self, jobId):
        db = database.getDB()
        dao = database.dao(db)
        try:
            if dao.getJobWithId(jobId) is None:
                loggerJobController.warn("job "+str(jobId)+" cannot be launched because it doesn't exist")
                return
            status = dao.getJobStatus(jobId)
            if status != database.jobStatus.DEACTIVATED or dao.canActivateJob(jobId):
                loggerJobController.info("activating job "+str(jobId))
                dao.activateJob(jobId)
                if dao.isGroupJob(jobId):
                    for jid in dao.getNoneGroupMembersRecursive(jobId):
                        self.jobs.add(jid)
                else:
                    self.jobs.add(jobId)
                self.processActiveJobs()
        finally:
            database.releaseDB(db)

    def startJob(self, jobId, responseId):
        self.queue.put(("start", jobId, responseId))

    def onStartJob(self, jobId, responseId):
        db = database.getDB()
        dao = database.dao(db)
        try:
            loggerJobController.info("Received message start job "+str(jobId))
            self.launchJob(jobId)
            if responseId is not None:
                dao.sendResponse(responseId, True, "ok")
        except Exception, e:
            loggerJobController.exception(str(e))
            if responseId is not None:
                dao.sendResponse(responseId, False, e.message)
            raise e
        finally:
            database.releaseDB(db)

    def _updateJobStatus(self, jobId, status, recursive, groupMembers):
        db = database.getDB()
        dao = database.dao(db)
        try:
            affectedJobs = set()
            job = dao.getJobWithId(jobId)
            if not job.groupJob:
                affectedJobs.add(jobId)
            if job.status != status:
                if job.groupJob and (status != database.jobStatus.FINISHED or job.status != database.jobStatus.RUNNING) and groupMembers:
                    for mId in dao.getGroupMemebers(jobId):
                        affectedJobs.add(mId)
                        self._updateJobStatus(mId, status, recursive, groupMembers)
                if status == database.jobStatus.IDLE or status == database.jobStatus.CANCELLED:
                    if not job.groupJob and job.status == database.jobStatus.RUNNING:
                        self.executer.terminate(job.executer, job.drmaaId)
                        dao.updateDRMAAId(jobId, None)
                    dao.updateJobStatus(jobId, status)
                elif status == database.jobStatus.FINISHED:
                    if job.status != database.jobStatus.RUNNING:
                        dao.updateJobStatus(jobId, status)
            if recursive:
                depJobIds = dao.getDependentJobs(jobId)
                for depId in depJobIds:
                    for affId in self._updateJobStatus(depId, status, recursive, True):
                        affectedJobs.add(affId)
            return affectedJobs
        finally:
            database.releaseDB(db)

    def respond(self, response_id, message):
        database.context(lambda db: database.dao(db).sendResponse(response_id, message))

    def run(self):
        loggerJobController.info("started job manager")
        db = database.getDB()
        dao = database.dao(db)
        running = dao.getRunningJobIds()
        dao.setRunningJobsToIdle()
        for id in running:
            j = dao.getJobWithId(id)
            if j.drmaaId is not None:
                self.executer.terminate(j.executer, j.drmaaId)
                dao.updateDRMAAId(id, None)
        self.jobs = set(dao.getActiveJobIds())
        self.processActiveJobs()
        database.releaseDB(db)
        while self.running:
            try:
                nr = 0
                while nr < 100:
                    nr += 1
                    req = self.queue.get(False)
                    messageType = req[0]
                    if messageType == "start":
                        self.onStartJob(req[1], req[2])
                    elif messageType == "update":
                        self.onUpdateStatus(req[1], req[2], req[3], req[4])
                    elif messageType == "delete":
                        self.onDeleteJob(req[1], req[2])
                    elif messageType == "stop":
                        loggerJobController.info("Received message stop")
                        self.running = False
            except Empty:
                pass
            except Exception, e:
                loggerJobController.exception(str(e))

            try:
                self.executer.handleMessages()
            except Exception, e:
                loggerJobController.exception(str(e))
            sleep(0.2)
        loggerJobController.info("stopped job manager")
        self.executer.stop()



def handleMessages(jobDAO, jobManager):
    transactionId = str(uuid4())
    try:
        messages = jobDAO.retreiveJobMessages(transactionId)
        for row in messages:
            messageType = row['type']
            if messageType == "launch":
                jobId = row['t_job_id']
                loggerMessageHandler.info("Received message start job "+str(jobId))
                jobManager.startJob(jobId, row['response_id'])
            elif messageType == "delete":
                jobId = row['t_job_id']
                loggerMessageHandler.info("Received message delete job "+str(jobId))
                jobManager.deleteJob(jobId, row['response_id'])
            elif messageType == "status":
                jobId = row['t_job_id']
                mess = row['message']
                vals = mess.split(":")
                status = vals[0]
                recursive = vals[1] == "True"
                loggerMessageHandler.info("Received message update job status (job_id, status, recursive): "+str(jobId)+", "+str(status)+", "+str(recursive))
                jobManager.updateStatus([jobId], status, recursive, row['response_id'])
    except:
        jobDAO.removeMessageTransactionId(transactionId)
    finally:
        jobDAO.deleteMessagesWithId(transactionId)

class MessageHandlerProcess(threading.Thread):

    def __init__(self, jobManager):
        threading.Thread.__init__(self)
        self.daemon = True
        self.jobManager = jobManager
        self.running = True

    def stop(self):
        self.running = False

    def run(self):
        loggerMessageHandler.info("Started database message handler")
        while self.running:
            db = database.getDB()
            try:
                dao = database.dao(db)
                handleMessages(dao, self.jobManager)
            except:
                traceback.print_exc()
            finally:
                database.releaseDB(db)
                time.sleep(5)
        logging.info("Database message handler finished")

class Jobber():

    def __init__(self):
        self.running = False


    def start(self):
        self.running = True
        if not os.path.exists(Settings.logDir):
            os.makedirs(Settings.logDir)
        logging.basicConfig(filename=Settings.logDir+"/main.log", format='%(asctime)s %(message)s', level=logging.INFO)
        db = database.getDB()
        db.init()
        jobDAO = database.dao(db)
        jobDAO.removeAllMessageTransactionIds()
        database.releaseDB(db)
        jobManager = JobManagerProcess()
        logging.info("starting job manager")
        jobManager.start()
        self.jobManager = jobManager
        messageHandler = MessageHandlerProcess(jobManager)
        logging.info("starting message handler")
        messageHandler.start()
        self.messageHandler = messageHandler

    def stop(self):
        logging.info("stopping job manager")
        self.jobManager.stop()
        self.messageHandler.stop()
        logging.info("stopping message handler")
        self.running = False



'''
Contains paths and settings
'''
import os

#if "JOBBER_CONFIG" not in os.environ:
#    sys.stderr.write("Could not find configuration. Please specify full path of configuration in environment variable JOBBER_CONFIG\n")
#    sys.exit(1)

drmaaMaxNrOfJobs = 1000

if "JOBBER_CONFIG" in os.environ:
    execfile(os.environ["JOBBER_CONFIG"])
else:
    execfile(os.path.join(os.path.expanduser("~"), ".jobber/config.py"))



from Queue import Queue, Empty
from subprocess import Popen
import threading
import time
from time import sleep
import os
import sys
import uuid

__author__ = 'rodak'

class BashJob():

    def __init__(self, stdout, stderr, script):
        self.script = script
        if stdout is not None:
            self.out = open(stdout, "w")
        else:
            self.out = None
        if stderr is not None:
            self.err = open(stderr, "w")
        else:
            self.err = None
        self.process = Popen([self.script], stdout=self.out, stderr=self.err, shell=True)

    def poll(self):
        return self.process.poll()

    def wait(self):
        return self.process.wait()

    def kill(self):
        self.process.kill()

EXECUTE=0
KILL=1
STOP=2

class ExecuterMessage():

    def __init__(self, type, message):
        self.type = type
        self.message = message

class RunJobMessage():

    def __init__(self, script, stdout, stderr, resultListener):
        self.script = script
        self.stdout = stdout
        self.stderr = stderr
        self.resultListener = resultListener
        self.maxRuntime = 10
        self.message = None
        self.jobId = None


RUNNING=0
FINISHED=1
TIMEOUT=2
KILLED=3

class RunJobResults():

    def __init__(self, status, returnCode, jobId):
        self.status = status
        self.returnCode = returnCode
        self.jobId = jobId

class SingleSystemCommandExecuter():

    def __init__(self, job, maxRuntime, listener, jobId):
        self.job = job
        self.jobId = jobId
        self.startTime = time.time()
        self.maxRuntime = maxRuntime
        self.listener = listener

    def checkJob(self):
        res = self.job.poll()
        if res is None:
            currentTime = time.time()
            runtime = currentTime - self.startTime
            if runtime > self.maxRuntime:
                try:
                    self.job.kill()
                    self.job.wait()
                except OSError:
                    pass
                self.listener(RunJobResults(TIMEOUT, None, self.jobId))
                return TIMEOUT
        else:
            self.listener(RunJobResults(FINISHED, res, self.jobId))
            return FINISHED
        return RUNNING

    def killJob(self):
        try:
            self.job.kill()
            self.job.wait()
        except OSError:
            pass
        self.listener(RunJobResults(KILLED, None, self.jobId))

def runOrDie(command):
    ret = os.system(command)
    if ret != 0:
        sys.stderr.write("Running command failed: "+command+"\n")
        sys.exit(ret)

class SystemCommandPooledExecuter(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self)
        self.poolSize = 6
        self.running = True
        self.queue = Queue()
        self.lock = threading.RLock()

    def execute(self, jobMessage):
        self.lock.acquire()
        try:
            jobId = str(uuid.uuid4())
            jobMessage.jobId = jobId
            self.queue.put(ExecuterMessage(EXECUTE, jobMessage))
            return jobId
        finally:
            self.lock.release()

    def kill(self, jobId):
        self.queue.put(ExecuterMessage(KILL, jobId))

    def run(self):
        idToExecuting = {}
        idToExecute = {}
        while self.running:
            try:
                message = self.queue.get(False)
                if message is not None:
                    if message.type == EXECUTE:
                        idToExecute[message.message.jobId] = message.message
                    elif message.type == KILL:
                        jobId = message.message
                        if jobId in idToExecute:
                            message = idToExecute[jobId]
                            del idToExecute[jobId]
                            message.resultListener(RunJobResults(KILLED, None, message.jobId))
                        elif jobId in idToExecuting:
                            ex = idToExecuting[jobId]
                            del idToExecuting[jobId]
                            ex.killJob()
                    elif message.type == STOP:
                        self.running = False
                        break
            except Empty:
                pass
            for jobId in idToExecuting.keys():
                ex = idToExecuting[jobId]
                j = ex.checkJob()
                if j != RUNNING:
                    del idToExecuting[jobId]
            while len(idToExecuting) < self.poolSize and len(idToExecute) > 0:
                key = idToExecute.iterkeys().next()
                message = idToExecute[key]
                del idToExecute[key]
                idToExecuting[key] = SingleSystemCommandExecuter(BashJob(message.stdout, message.stderr, message.script), message.maxRuntime, message.resultListener, key)
            sleep(0.3)


    def stop(self):
        self.running = False
        self.queue.put(ExecuterMessage(STOP, None))



'''
Created on Jun 3, 2014

@author: rodak
'''
import os
import shutil
from Jobber import Settings
from Jobber.database import JobStatus, JobAction


class Option():
    
    def __init__(self, name, value):
        self.name = name
        self.value = value

class Dependency():
    def __init__(self, dependency_id, dependency_status):
        self.dependencyId = dependency_id
        self.dependencyStatus = dependency_status
        
class NameAndParentName():
    def __init__(self, name, parentName = None):
        self.name = name
        self.parentName = parentName

    def equal(self, nameAndParentName):
        thisName = self
        otherName = nameAndParentName
        while thisName.name == otherName.name:
            if thisName.parentName is None:
                return otherName.parentName is None
            if otherName.parentName is None:
                return False
            thisName = thisName.parentName
            otherName = otherName.parentName
        return False


class PersistedJob():
    
    def __init__(self):
        self.dependencies = []

class JobDAO():
    
    def __init__(self, db):
        self.db = db

    def _booleanToYN(self, b):
        if b:
            return "Y"
        return "N"

    ################ Methods required by web interface  ###############################

    def getPipelines(self):
        self.db.execute("select * from t_jobs where parent_id is NULL order by id desc")
        return self.db.fetchall()

    def getMembers(self, JobID):
        self.db.execute("select * from t_jobs where parent_id = %i order by id" % JobID)
        return self.db.fetchall()

    def updateCommand(self, JobID, new_command):
        self.db.execute("update t_jobs set job_command=\"%s\" where id = %i;" % (new_command, JobID))

    ################ Methods required by web interface  ###############################

    def getActiveJobs(self):
        res = []
        self.db.execute("SELECT * FROM t_jobs WHERE status = 'IDLE' and is_group_job = 'N'")
        for row in self.db.fetchall():
            job = self._rowToJob(row)
            res.append(job)
        return res

    def sendLaunchJobMessage(self, jobId, responseId = None):
        self.db.execute("INSERT INTO t_job_messages (t_job_id, type, response_id) VALUES (%s, 'launch', %s)", [jobId, responseId])

    def sendDeleteJobMessage(self, jobId, responseId = None):
        self.db.execute("INSERT INTO t_job_messages (t_job_id, type, response_id) VALUES (%s, 'delete', %s)", [jobId, responseId])

    def sendChangeStatusJobMessage(self, jobId, status, recursive = False, response_id = None):
        self.db.execute("INSERT INTO t_job_messages (t_job_id, type, message, response_id) VALUES (%s, 'status', %s, %s)", [jobId, status+":"+str(recursive), response_id])

    def retreiveJobMessages(self, handlerId):
        self.db.execute("UPDATE t_job_messages SET transaction_id = %s WHERE transaction_id is null", [handlerId])
        self.db.execute("SELECT * FROM t_job_messages WHERE transaction_id = %s", [handlerId])
        return self.db.fetchall()

    def sendResponse(self, responseId, success, message):
        dbSuccess = 0
        if success:
            dbSuccess = 1
        self.db.execute("INSERT INTO t_job_responses (id, success, message) VALUES (%s, %s, %s)", [responseId, dbSuccess, message]);

    def removeMessageTransactionId(self, trId):
        self.db.execute("UPDATE t_job_messages SET transaction_id = NULL WHERE transaction_id = %s", [trId])

    def removeAllMessageTransactionIds(self):
        self.db.execute("UPDATE t_job_messages SET transaction_id = NULL")

    def deleteMessagesWithId(self, trId):
        self.db.execute("DELETE FROM t_job_messages WHERE transaction_id = %s", [trId])

    def getActiveJobIds(self):
        res = []
        self.db.execute("SELECT id FROM t_jobs WHERE status = 'IDLE' and is_group_job = 'N'")
        for row in self.db.fetchall():
            res.append(row['id'])
        return res

    def getRunningJobIds(self):
        res = []
        self.db.execute("SELECT id FROM t_jobs WHERE status = 'RUNNING' and is_group_job = 'N'")
        for row in self.db.fetchall():
            res.append(row['id'])
        return res

    def getJobWithDrmaaId(self, drmaaId):
        self.db.execute("SELECT * FROM t_jobs WHERE drmaa_id = %s", [drmaaId])
        res = self.db.fetchone()
        if res is not None:
            res = self._rowToJob(res)
        return res

    def setRunningJobsToIdle(self):
        self.db.execute("UPDATE t_jobs SET status = 'IDLE' WHERE status = 'RUNNING' and is_group_job = 'N'")

    def getTotalRuntime(self, jobId):
        time = 0
        if self.isGroupJob(jobId):
            for jid in self.getGroupMemebers(jobId):
                time += self.getTotalRuntime(jid)
        else:
            jobInfo = self.getJobInfo(jobId)
            time = (jobInfo['end_run_date'] - jobInfo['start_run_date']).total_seconds()
        return time

    def persistNewJob(self, job):
        self.db.execute("""INSERT INTO t_jobs(name, parent_id, description, status, job_command, is_unique,
            unique_key, is_group_job, max_parallel_nr, max_nr_of_restarts, current_run, delete_time, who_create, who_update, executer)
                               VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)""",
                               [job.name, job.parentId, job.description, job.status, job.command, self._booleanToYN(job.unique), job.uniqueKey,
                                self._booleanToYN(job.groupJob), job.maxParallelNr, job.maxNrOfRestarts, job.currentRun, job.deleteTime,
                                job.whoCreate, job.whoUpdate, job.executer])
        job.jobId = self.db.insert_id()
        self.persistJobDependencies(job)
        self.persistJobOptions(job)
        return job.jobId

    def updateStartSubmitDate(self, jobId, date):
        self.db.execute("UPDATE t_jobs SET start_submit_date = %s WHERE id = %s", [date, jobId])

    def updateEndSubmitDate(self, jobId, date):
        self.db.execute("UPDATE t_jobs SET end_submit_date = %s WHERE id = %s", [date, jobId])

    def updateStartRunDate(self, jobId, date):
        self.db.execute("UPDATE t_jobs SET start_run_date = %s WHERE id = %s", [date, jobId])

    def updateEndRunDate(self, jobId, date):
        self.db.execute("UPDATE t_jobs SET end_run_date = %s WHERE id = %s", [date, jobId])

    def updateDRMAAId(self, jobId, drmaaId):
        self.db.execute("UPDATE t_jobs set drmaa_id = %s WHERE id = "+str(jobId), [drmaaId])

    def updateError(self, jobId, error):
        self.db.execute("UPDATE t_jobs set error = %s WHERE id = %s", [error, jobId])

    def deleteJob(self, jobId):
        if self.isGroupJob(jobId):
            members = self.getGroupMemebers(jobId)
            for member in members:
                self.deleteJob(member)
        else:
            logDir = Settings.logDir+"/"+str(jobId)
            if os.path.exists(logDir):
                shutil.rmtree(logDir)
        self.db.execute("DELETE FROM t_jobs WHERE id = %s", [jobId])
        self.db.execute("DELETE FROM t_job_dependencies WHERE t_job_id = %s", [jobId])
        self.db.execute("DELETE FROM t_job_options WHERE t_job_id = %s", [jobId])

    def getExistingUniqueJobId(self, job):
        if job.unique:
            uniqueKey = job.uniqueKey
            parentId = job.parentId
            args = []
            command = "SELECT * FROM t_jobs WHERE "
            if parentId is not None:
                command += "parent_id = %s AND "
                args.append(parentId)
            if uniqueKey is None:
                uniqueKey = job.command
                command += " job_command = %s"
            else:
                command += " unique_key = %s"
            args.append(uniqueKey)
            self.db.execute(command, args)
            job = self.db.fetchone()
            if job is None:
                return job
            return self._rowToJob(job)

    def getJobInfo(self, jobId):
        command = "SELECT * FROM t_jobs WHERE id = %s"
        self.db.execute(command, [jobId])
        return self.db.fetchone()

    def isRetryJob(self, jobId):
        job = self.getJobWithId(jobId)
        if job.currentRun < job.maxNrOfRestarts:
            return True
        return False

    def canActivateJob(self, jobId):
        jinfo = self.getJobInfo(jobId)
        if (('t_parent_id' not in jinfo) or jinfo['t_parent_id'] is None) and jinfo['max_parallel_nr'] > 0:
            self.db.execute("SELECT count(*) as c FROM t_jobs WHERE status = 'RUNNING' and name = '"+jinfo['name']+"'")
            running = self.db.fetchone()['c']
            return jinfo['max_parallel_nr'] > running
        else:
            return True

    def activateJob(self, jobId):
        if self.isGroupJob(jobId):
            for jid in self.getGroupMemebers(jobId):
                self.activateJob(jid)
        if self.getJobStatus(jobId) == JobStatus.DEACTIVATED:
            self.updateJobStatus(jobId, JobStatus.IDLE)

    def getActivableJob(self, name):
        self.db.execute("SELECT id, max_parallel_nr FROM t_jobs WHERE status = 'SUSPENDED' AND name= '"+name+"' ORDER BY id LIMIT 1")
        res = self.db.fetchone()
        if res is not None:
            return res['id']
        return None


    def persistJobDependencies(self, job):
        for dependency in job.dependencies:
            self.db.execute("""
                    INSERT INTO t_job_dependencies(t_job_id, t_job_dependency_id, dependency_job_status) VALUES (%s, %s, %s)
                """, [job.jobId, dependency.dependencyId, dependency.dependencyStatus])
    
    def persistJobOptions(self, job):
        for option in job.options:
            self.db.execute("""
                    INSERT INTO t_job_options (t_job_id, option_name, option_value) VALUES (%s, %s, %s)
                """, [job.jobId, option.name, option.value])

    def isGroupJob(self, jobId):
        self.db.execute("SELECT is_group_job FROM t_jobs WHERE id = %s", [jobId])
        gj = self.db.fetchone()['is_group_job']
        return gj == "Y"

    def getDependentJobs(self, jobId):
        self.db.execute("SELECT t_job_id FROM t_job_dependencies WHERE t_job_dependency_id = %s", [jobId])
        r = []
        for row in self.db.fetchall():
            r.append(row['t_job_id'])
        return r

    def getAffectedJobs(self, jobId):
        deps = self.getDependentJobs(jobId)
        res = []
        for dep in deps:
            if self.isGroupJob(dep):
                res += self.getNoneGroupMembersRecursive(dep)
            else:
                res.append(dep)
        return res

    def getParentId(self, jobId):
        self.db.execute("SELECT parent_id FROM t_jobs WHERE id = %s", [jobId])
        f = self.db.fetchone()
        if f is None:
            return f
        return f['parent_id']
        
    def updateJobStatus(self, jobId, status):
        self.db.execute("UPDATE t_jobs SET status = '"+status+"' WHERE id = "+str(jobId))

    def updateRun(self, jobId):
        self.db.execute("UPDATE t_jobs SET current_run = current_run + 1 WHERE id = "+str(jobId))
    
    def getJobWithId(self, jobId):
        self.db.execute("SELECT * FROM t_jobs WHERE id = "+str(jobId))
        for row in self.db.fetchall():
            return self._rowToJob(row)
        return None

    def getGroupMemebers(self, jobId):
        self.db.execute("SELECT id FROM t_jobs WHERE parent_id = "+str(jobId))
        res = []
        for row in self.db.fetchall():
            res.append(row['id'])
        return res

    def getNoneGroupMembers(self, jobId):
        self.db.execute("SELECT id FROM t_jobs WHERE parent_id = "+str(jobId)+" and is_group_job = 'N'")
        res = []
        for row in self.db.fetchall():
            res.append(row['id'])
        return res

    def getNoneGroupMembersRecursive(self, jobId):
        members = self.getGroupMemebers(jobId)
        res = []
        for jid in members:
            if self.isGroupJob(jid):
                res += self.getNoneGroupMembersRecursive(jid)
            else:
                res.append(jid)
        return res

    def isGroupJobFailed(self, groupJobId):
        self.db.execute("SELECT count(*) as c from t_jobs where parent_id = %s AND status = 'FAILED'", [groupJobId])
        return self.db.fetchone()['c'] > 0

    def isGroupJobCancelled(self, groupJobId):
        self.db.execute("SELECT count(*) as c from t_jobs where parent_id = %s AND status = 'CANCELLED'", [groupJobId])
        return self.db.fetchone()['c'] > 0

    def isGroupJobRunning(self, groupJobId):
        self.db.execute("SELECT count(*) as c from t_jobs where parent_id = %s AND status = 'RUNNING'", [groupJobId])
        return self.db.fetchone()['c'] > 0

    def isGroupJobFinished(self, groupJobId):
        self.db.execute("select count(*) as c from t_jobs where parent_id = %s", [groupJobId])
        total = self.db.fetchone()['c']
        self.db.execute("select count(*) as c from t_jobs where parent_id = %s and status = 'FINISHED'", [groupJobId])
        totalFinished = self.db.fetchone()['c']
        return total == totalFinished
    
    def getJobDependencyIds(self, jobId):
        self.db.execute("SELECT t_job_dependency_id, dependency_job_status FROM t_job_dependencies WHERE t_job_id = "+str(jobId))
        res = []
        for row in self.db.fetchall():
            res.append(Dependency(row['t_job_dependency_id'], row['dependency_job_status']))
        return res

    def getParentJobDependencyIds(self, jobId):
        parentId = self.getParentId(jobId)
        if parentId is None:
            return []
        res = self.getJobDependencyIds(parentId)
        return res + self.getParentJobDependencyIds(parentId)

    def getJobStatus(self, jobId):
        self.db.execute("SELECT status FROM t_jobs WHERE id = (%s)", [jobId])
        return self.db.fetchone()['status']

    def getGroupJobAction(self, jobId):
        jobStatus = self.getJobStatus(jobId)
        if self.isGroupJobFailed(jobId):
            return JobAction.NOTHING if jobStatus == JobStatus.FAILED else JobAction.FAIL
        if self.isGroupJobCancelled(jobId):
            return JobAction.CANCEL if jobStatus != JobStatus.CANCELLED else JobAction.NOTHING
        if self.isGroupJobFinished(jobId):
            return JobAction.FINISH_GROUP if jobStatus != JobStatus.FINISHED else JobAction.NOTHING
        if self.isGroupJobRunning(jobId):
            return JobAction.EXECUTE if jobStatus != JobStatus.RUNNING else JobAction.NOTHING
        else:
            return JobAction.IDLE_GROUP if jobStatus != JobStatus.IDLE else JobAction.NOTHING

    def getNameWithParentNames(self, jobId):
        jobInfo = self.getJobInfo(jobId)
        name = jobInfo['name']
        parent = jobInfo['parent_id']
        parentName = None
        if parent is not None:
            parentName = self.getNameWithParentNames(parent)
        return NameAndParentName(name, parentName)

    def getIdleMaxParallelRunningJobsWithParentAndParentNames(self, nameAndParentNames):
        name = nameAndParentNames.name
        self.db.execute("SELECT id FROM t_jobs WHERE status = 'IDLE' AND name = %s and max_parallel_nr > 0", [name])
        res = []
        for row in self.db.fetchall():
            jobId = row['id']
            npn = self.getNameWithParentNames(jobId)
            if npn.equal(nameAndParentNames):
                res.append(jobId)
        return res

    def getRunningJobsWithNameAndParentNames(self, nameAndParentNames):
        name = nameAndParentNames.name
        self.db.execute("SELECT id FROM t_jobs WHERE status = 'RUNNING' AND name = %s", [name])
        res = []
        for row in self.db.fetchall():
            jobId = row['id']
            npn = self.getNameWithParentNames(jobId)
            if npn.equal(nameAndParentNames):
                res.append(jobId)
        return res

    def getMaxNrOfRunning(self, jobId):
        self.db.execute("SELECT max_parallel_nr FROM t_jobs WHERE id = %s", [jobId])
        row = self.db.fetchone()
        running = row['max_parallel_nr']
        return running

    def getJobAction(self, jobId):
        jobStatus =  self.getJobStatus(jobId)
        if jobStatus != JobStatus.IDLE and jobStatus != JobStatus.RUNNING:
            return JobAction.NOTHING
        deps = self.getJobDependencyIds(jobId) + self.getParentJobDependencyIds(jobId)
        res = JobAction.EXECUTE
        for dep in deps:
            depStat = dep.dependencyStatus
            stat = self.getJobStatus(dep.dependencyId)
            if depStat != stat:
                return JobAction.NOTHING
        if res == JobAction.EXECUTE:
            if jobStatus != JobStatus.IDLE:
                return JobAction.NOTHING
        return res

    def getJobIdsWithFinishedDirectDependencies(self):
        self.db.execute("c")

    def getFailureReason(self):
        return self.failReason

    def getOptions(self, jobId):
        parentJobId = self.getParentId(jobId)
        parentOptions = []
        if parentJobId is not None:
            parentOptions = self.getOptions(parentJobId)
        self.db.execute("SELECT option_name, option_value FROM t_job_options WHERE t_job_id = "+str(jobId))
        res = []
        for pOpt in parentOptions:
            res.append(pOpt)
        for row in self.db.fetchall():
            res.append(Option(row['option_name'], row['option_value']))
        return res

    def getExecuter(self, jobId):
        self.db.execute("SELECT executer FROM t_jobs WHERE id = "+str(jobId))
        row = self.db.fetchone()
        if row is None:
            return Settings.defaultExecuter
        else:
            if row['executer'] is not None:
                return row['executer']
            else:
                parent = self.getParentId(jobId)
                if parent is not None:
                    return self.getExecuter(parent)
                return Settings.defaultExecuter


    def _rowToJob(self, row):
        job = PersistedJob()
        job.jobId = row['id']
        isGroup = row['is_group_job']
        if 'name' in row:
            job.name = row['name']
        else:
            job.name = None
        if 'description' in row:
            job.description = row['description']
        else:
            job.description = None
        job.status = row['status']
        job.unique = row['is_unique'] == 'Y'
        if 'unique_key' in row:
            job.uniqueKey = row['unique_key']
        else:
            job.uniqueKey = None
        if 'startDate' in row:
            job.startDate = row['start_date']
        else:
            job.startDate = None
        if 'endDate' in row:
            job.endDate = row['end_date']
        else:
            job.endDate = None
        if 'delete_time' in row:
            job.deleteTime = row['delete_time']
        else:
            job.deleteTime = None
        if 'who_create' in row:
            job.whoCreate = row['who_create']
        else:
            job.whoCreate = None
        if 'who_update' in row:
            job.whoUpdate = row['who_update']
        else:
            job.whoUpdate = None
        if 'parent_id' in row:
            job.parentId = row['parent_id'];
        else:
            job.parentId = None
        job.maxParallelNr = row['max_parallel_nr']
        job.maxNrOfRestarts = row['max_nr_of_restarts']
        job.currentRun = row['current_run']
        job.executer = self.getExecuter(job.jobId)
        job.dependencies = self.getJobDependencyIds(job.jobId)
        job.options = self.getOptions(job.jobId)
        if isGroup == 'Y':
            job.groupJob = True
            job.command = None
            job.drmaaId = None
        else:
            job.groupJob = False

            if 'drmaa_id' in row:
                job.drmaaId = row['drmaa_id']
            else:
                job.drmaaId = None
            job.command = row['job_command']
        return job
        



from . import database

__author__ = 'rodak'

def is_sequence(arg):
    return (not hasattr(arg, "strip") and
            hasattr(arg, "__getitem__") or
            hasattr(arg, "__iter__"))

class Jobber():

    def __init__(self):
        self.jobDAO = database.dao(database.getDB())
        self.parentIds = []
        self.jobCreator = database.JobCreator()
        self.ignoreJobId = None

    def _setSettings(self, settings):
        for set in settings:
            if set == "name":
                self.jobCreator.name = settings[set]
            elif set == "dependencies":
                for dep in settings[set]:
                    stat = "FINISHED"
                    if not is_sequence(dep):
                        jobId = dep
                    else:
                        jobId = dep[0]
                        if len(dep) > 1:
                            stat = dep[1]
                    self.jobCreator.dependencies.append(database.JobDependency(long(jobId), stat))
            elif set == "options":
                for opt in settings[set]:
                    self.jobCreator.options.append(database.JobOption(opt[0], opt[1]))
            elif set == "description":
                self.jobCreator.description = settings[set]
            elif set == "name":
                self.jobCreator.name = settings[set]
            elif set == "executer":
                self.jobCreator.executer = settings[set]
            elif set == "maxNrOfJobs":
                self.jobCreator.maxNrOfParallelJobs = int(settings[set])
            elif set == "uniqueId":
                self.jobCreator.uniqueJob = True
                uId = settings[set]
                if isinstance(uId, basestring):
                    self.jobCreator.uniqueKey = uId

    def job(self, command, settings = {}):
        if self.ignoreJobId is not None:
            return -1
        self._setSettings(settings)
        job = self.jobCreator.createJob(command)
        if len(self.parentIds) > 0:
            job.parentId = self.parentIds[len(self.parentIds) - 1]
        existingJob = self.jobDAO.getExistingUniqueJobId(job)
        if existingJob is not None:
            return existingJob.jobId
        else:
            newJob = self.jobDAO.persistNewJob(job)
            return newJob

    def startGroup(self, settings = {}):
        if self.ignoreJobId is not None:
            self.parentIds.append(-1)
            return -1
        self._setSettings(settings)
        self.jobCreator.groupJob = True
        job = self.jobCreator.createJob("GROUP")
        if len(self.parentIds) > 0:
            job.parentId = self.parentIds[len(self.parentIds) - 1]
        existingJob = self.jobDAO.getExistingUniqueJobId(job)
        if existingJob is not None:
            self.ignoreJobId =  existingJob.jobId
            self.parentIds.append( existingJob.jobId)
            return  existingJob.jobId
        else:
            newJob = self.jobDAO.persistNewJob(job)
            self.parentIds.append(newJob)
            return newJob

    def extendGroup(self, groupId):
        self.parentIds.append(groupId)

    def endGroup(self):
        jobId = self.parentIds.pop()
        if self.ignoreJobId is not None:
            if self.ignoreJobId == jobId:
                self.ignoreJobId = None

    def launch(self, jobId):
        self.jobDAO.sendLaunchJobMessage(jobId)

    def delete(self, jobId):
        self.jobDAO.sendDeleteJobMessage(jobId)

    def changeStatus(self, jobId, status, recursive = False, resonse_id = None):
        self.jobDAO.sendChangeStatusJobMessage(jobId, status, recursive, resonse_id)

    def close(self):
        database.releaseDB(self.jobDAO.db)

JobberBuilder = Jobber



'''
Created on Mar 24, 2010

@author: rodak

'''
from Jobber import Settings

from . import Mysql
from . import Sqlite3


def createMysqlDB():
    return Mysql.DB()

def createSqliteDB():
    return Sqlite3.DB()

try:
    sqlitePath = Settings.sqlite_path
    createDB = createSqliteDB
except AttributeError:
    createDB = createMysqlDB


#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(name='Jobber',
      version='0.1.1',
        description=("Jobber is a program that executes pipelines. It manages job dependencies,"
                     " restarts them if necessary and logs their status in a database. Jobs are"
                     " persisted in a database and survive system crashes."),
      long_description=readme + '\n\n' + history,
      author="Christoph Rodak",
      author_email='christoph@rodak.li',
      url='https://git.scicore.unibas.ch/AnnotationPipelines/Jobber',
      packages=[
          'Jobber',
      ],
      package_dir={'Jobber':
                   'Jobber'},
      include_package_data=True,
      install_requires=requirements,
      license="BSD",
      zip_safe=False,
      keywords='Jobber',
      scripts=['scripts/jobber_server', 'scripts/jobber_web'],
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        # "Programming Language :: Python :: 2",
        # 'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.3',
        # 'Programming Language :: Python :: 3.4',
      ],
      test_suite='tests',
      tests_require=test_requirements
)
